//
// LED_neopixel.c
//
// NeoPixel driving for the 12th Doctor's second Sonic Screwdriver
//
#include <Arduino.h>
#include "LED_neopixel.h"

extern void timer0_off();          // turn timer 0 interrupt off
extern void timer0_on();           // turn timer 0 interrupt on
extern uint8_t fastISR;           // Sound generation ISR speed switch

//---------------------------------------------------------------------------------------------------
//
//                                            LED driving
//
/****************************************************************************************************
 This is an example of how simple driving a Neopixel can be
 This code is optimized for understandability and changability rather than raw speed
 More info at http://wp.josh.com/2014/05/11/ws2812-neopixels-made-easy/

 *!* very heavily modified by TonyWilk *!*
*/

#define PIXEL_PORT  PORTD  // Port of the pin the pixels are connected to
#define PIXEL_DDR   DDRD   // Port of the pin the pixels are connected to
#define PIXEL_BIT   4      // Bit of the pin the pixels are connected to
  
// These are the timing constraints taken mostly from the WS2812 datasheets 
// These are chosen to be conservative and avoid problems rather than for maximum throughput 

#define T1H  1100    // Width of a 1 bit in ns
#define T1L  350

#define T0H  350    // Width of a 0 bit in ns
#define T0L  1360

#define RES 60000    // Width of the low gap between bits to cause a frame to latch

#define CYCLES_PER_MS (F_CPU / 1000L)
#define NS_PER_MS (1000000L)
#define NS_TO_CYCLES(n) ( ((n) * CYCLES_PER_MS) / NS_PER_MS )

// Set the specified pin up as digital out
//
void ledSetup() {
  bitSet( PIXEL_DDR , PIXEL_BIT );
}

// This has now been reduced to the minimum timing for a 16MHz ATMEGA by TonyWilk
// - the big problem is having to allow the Timer2 interrupt to generate sound
//   while we're outputting the LED data... if there's a gap of more than about 6uS, the LED's reset
//   So...
//   1. high periods are timed with interrupts disabled
//   2. we reduce the low periods as much as possible and allow interrupts
//   3. the Timer2 ISR in the main code has a 'fastISR' mode to take as little time as possible
//   4. we disable the Timer0 interrupts for the duration of the LED drive
//   5. LEDs will flicker if you use Serial.print and immediately send LED data (Serail uses interrupt)

inline void sendBit( uint8_t bitVal ) 
{
  if (  bitVal ) {        // 1 bit
      
      asm volatile (
        "cli \n\t"                        // INTERRUPTS OFF for the high period
        "sbi %[port], %[bit] \n\t"        // Set the output bit
        ".rept %[onCycles] \n\t"          // Execute NOPs to delay exactly the specified number of cycles
        "nop \n\t"
        ".endr \n\t"
        "cbi %[port], %[bit] \n\t"        // Clear the output bit
        "sei \n\t"
        //      ".rept %[offCycles] \n\t"         // Execute NOPs to delay exactly the specified number of cycles
        "nop \n"                          // just a single nop will do
        //      ".endr \n\t"
        ::
        [port]    "I" (_SFR_IO_ADDR(PIXEL_PORT)),
        [bit]   "I" (PIXEL_BIT),
        [onCycles]  "I" (NS_TO_CYCLES(T1H) - 2),    // 1-bit width less overhead  for the actual bit setting, note that this delay could be longer and everything would still work
        [offCycles]   "I" (NS_TO_CYCLES(T1L) - 2)     // Minimum interbit delay. Note that we probably don't need this at all since the loop overhead will be enough, but here for correctness
  
      );
  } else {          // 0 bit

      // **************************************************************************
      // This line is really the only tight goldilocks timing in the whole program!
      // **************************************************************************
  
      asm volatile (
        "cli \n\t"                      // INTERRUPTS OFF for the high period
        "sbi %[port], %[bit] \n\t"      // Set the output bit
        ".rept %[onCycles] \n\t"        // Now timing actually matters. The 0-bit must be long enough to be detected but not too long or it will be a 1-bit
        "nop \n\t"                      // Execute NOPs to delay exactly the specified number of cycles
        ".endr \n\t"
        "cbi %[port], %[bit] \n\t"      // Clear the output bit
        "sei \n\t"
  
        //      ".rept %[offCycles] \n\t"       // Execute NOPs to delay exactly the specified number of cycles
        //      "nop \n"
        //      "nop \n"
        //      "nop \n"             // no delay needed due to function call overhead
        //      "nop \n"
        //      "nop \n"
        //      ".endr \n\t"
        ::
        [port]    "I" (_SFR_IO_ADDR(PIXEL_PORT)),
        [bit]   "I" (PIXEL_BIT),
        [onCycles]  "I" (NS_TO_CYCLES(T0H) - 2),
        [offCycles] "I" (NS_TO_CYCLES(T0L) - 2)
      );
  }
}  

inline void sendByte( uint8_t val )
{
    sendBit( val & 0x80 );     // inlined to avoid further inter-bit gaps
    sendBit( val & 0x40 );
    sendBit( val & 0x20 );
    sendBit( val & 0x10 );
    sendBit( val & 0x08 );
    sendBit( val & 0x04 );
    sendBit( val & 0x02 );
    sendBit( val & 0x01 );
} 

//------------------------------------------------------------------
// sendLeds
// - sends pixels given pallette colours and brightness in 'colours'
// - scales values by 'globalBrightness'
//
void sendLeds( struct LED_PIXEL *leds, struct LED_RGB *pixelBuf, uint8_t brightness, uint8_t nPixels )
{
  uint16_t i, colour, index, bright;

  // prepare buffer of pixel RGB values
  for( index=i=0; i<nPixels; i++ ){
    colour= leds[i].colour;
    
    bright= ((uint16_t)leds[i].bright * brightness) >> 8;
    bright= pgm_read_byte_near( gammaTab + bright );
    pixelBuf[i].r= ((uint16_t)pgm_read_byte_near(pal_red + colour) * bright) >> 8;
    pixelBuf[i].g= ((uint16_t)pgm_read_byte_near(pal_green + colour) * bright) >> 8;
    pixelBuf[i].b= ((uint16_t)pgm_read_byte_near(pal_blue + colour) * bright) >> 8;
  }

  cli();
  fastISR= 1;     // enable fastISR mode for Timer2 sound output ISR
  timer0_off();   // disable timer0 interrupt (used for delay()) 
  sei();
  
  sendByte( pixelBuf[0].r );    // output LED string
  sendByte( pixelBuf[0].g );    // inlined to avoid extra gaps between bits
  sendByte( pixelBuf[0].b );    // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  sendByte( pixelBuf[1].r );
  sendByte( pixelBuf[1].g );
  sendByte( pixelBuf[1].b );
  sendByte( pixelBuf[2].r );
  sendByte( pixelBuf[2].g );
  sendByte( pixelBuf[2].b );
  sendByte( pixelBuf[3].r );
  sendByte( pixelBuf[3].g );
  sendByte( pixelBuf[3].b );
  sendByte( pixelBuf[4].r );
  sendByte( pixelBuf[4].g );
  sendByte( pixelBuf[4].b );
  
  cli();
  fastISR= 0;
  timer0_on();
  sei();
  //delayMicroseconds( 50 );   // delay to latch LEDs - not reqd. we don't call this that fast
}

/* end of LED_neopixel.cpp */

