//
// LSM9DS0 accelerometer,gyro,magnetometer (on FLORA 9-DOF board)
//
/****************************************************************/
#ifndef __ACCEL_H__
	#define __ACCEL_H__
/****************************************************************/

#include <ci2c.h>
#include <math.h>

#ifdef __cplusplus
extern "C"{
#endif


#define DEGREES_PER_RADIAN (180.0/M_PI)
#define ACCEL_SCALE 0.598          // accel mm/s^2 per bit at +-2G scale
#define GYRO_SCALE  0.00875        // rotation degrees/sec at +-245dps scale
#define GRAVITY 9800               // gravity in mm/s^2
#define MAG_SCALE   0.08           // mgauss

// The LSM9DS0 has two I2C slave addresses:
//    0x3A for acceleration and mag. sensor (table 15 in data sheet)
//    0xD6 for control regs, Gyro, temp etc.     (table 16)
//
#define ACCEL_ADDR (0x3A >> 1) 
#define GYRO_ADDR  (0xD6 >> 1)     

// XYZ type
//
union __attribute__((__packed__)) HILO{
  int16_t val;
  struct{
    uint8_t lo;
    uint8_t hi;
  }hilo;
};

struct XYZ{
  union HILO X;
  union HILO Y; 
  union HILO Z;
};


struct VECXYZ {
  int x;
  int y;
  int z;
};


//--------------------------------------------------------------------------------
//                                     Functions
//--------------------------------------------------------------------------------
I2C_STATUS accel_init();

// Read acceleration (raw values from device);
//
void readAccel( struct XYZ *xyz );

// Read Gyro (raw values from device)
//
void readGyro( struct XYZ *xyz);

// Read magnetometer (raw values from device)
//
void readMag( struct XYZ *xyz );

// Calculate heading in degrees from raw X,Y magnetometer values
// returns 0 .. +- 180
//
int calcHeading( struct XYZ *rawMag );


// processAccel
// - this is the API to read the accelerometer
// - outputs:
//      axisRaw = raw acceleration along each axis (incl. gravity) scaled to mm/sec^2
//      axisAccel = accelerations along each axis (minus gravity) in mm/sec^2
//      axisDegrees = tilt of each axis in integer degrees (0=level +90 = up, -90 = down)
//
void processAccel( struct VECXYZ * axisRaw, struct VECXYZ * axisAccel, struct VECXYZ * axisDegrees );

//
// Magnetometer calibration
//
#define MAG_CAL_OFF		0
#define MAG_CAL_INIT	1
#define MAG_CAL_RUN		2
#define MAG_CAL_STOP	3
void setMagCalMode( uint8_t mode );
void processMagCal( struct XYZ *xyz );
void getMagCal( struct XYZ *xyz );

// Internal functions:

// read single byte from register
//
I2C_STATUS read8( uint8_t slaveAddr, uint8_t regAddr, uint8_t *dest );

// write single byte to register
//
I2C_STATUS write8( uint8_t slaveAddr, uint8_t regAddr, uint8_t data );


//---------------------------------------------------
// Device registers etc.... (from Adafruit_LSM9DS0.h)
//
typedef enum
{
  LSM9DS0_REGISTER_WHO_AM_I_G          = 0x0F,
  LSM9DS0_REGISTER_CTRL_REG1_G         = 0x20,
  LSM9DS0_REGISTER_CTRL_REG3_G         = 0x22,
  LSM9DS0_REGISTER_CTRL_REG4_G         = 0x23,
  LSM9DS0_REGISTER_OUT_X_L_G           = 0x28,
  LSM9DS0_REGISTER_OUT_X_H_G           = 0x29,
  LSM9DS0_REGISTER_OUT_Y_L_G           = 0x2A,
  LSM9DS0_REGISTER_OUT_Y_H_G           = 0x2B,
  LSM9DS0_REGISTER_OUT_Z_L_G           = 0x2C,
  LSM9DS0_REGISTER_OUT_Z_H_G           = 0x2D
} lsm9ds0GyroRegisters_t;

typedef enum
{
  LSM9DS0_REGISTER_TEMP_OUT_L_XM       = 0x05,
  LSM9DS0_REGISTER_TEMP_OUT_H_XM       = 0x06,
  LSM9DS0_REGISTER_STATUS_REG_M        = 0x07,
  LSM9DS0_REGISTER_OUT_X_L_M           = 0x08,
  LSM9DS0_REGISTER_OUT_X_H_M           = 0x09,
  LSM9DS0_REGISTER_OUT_Y_L_M           = 0x0A,
  LSM9DS0_REGISTER_OUT_Y_H_M           = 0x0B,
  LSM9DS0_REGISTER_OUT_Z_L_M           = 0x0C,
  LSM9DS0_REGISTER_OUT_Z_H_M           = 0x0D,
  LSM9DS0_REGISTER_WHO_AM_I_XM         = 0x0F,
  LSM9DS0_REGISTER_INT_CTRL_REG_M      = 0x12,
  LSM9DS0_REGISTER_INT_SRC_REG_M       = 0x13,
  LSM9DS0_REGISTER_CTRL_REG1_XM        = 0x20,
  LSM9DS0_REGISTER_CTRL_REG2_XM        = 0x21,
  LSM9DS0_REGISTER_CTRL_REG5_XM        = 0x24,
  LSM9DS0_REGISTER_CTRL_REG6_XM        = 0x25,
  LSM9DS0_REGISTER_CTRL_REG7_XM        = 0x26,
  LSM9DS0_REGISTER_OUT_X_L_A           = 0x28,
  LSM9DS0_REGISTER_OUT_X_H_A           = 0x29,
  LSM9DS0_REGISTER_OUT_Y_L_A           = 0x2A,
  LSM9DS0_REGISTER_OUT_Y_H_A           = 0x2B,
  LSM9DS0_REGISTER_OUT_Z_L_A           = 0x2C,
  LSM9DS0_REGISTER_OUT_Z_H_A           = 0x2D
} lsm9ds0MagAccelRegisters_t;

typedef enum
{
  LSM9DS0_ACCELRANGE_2G                = (0b000 << 3),
  LSM9DS0_ACCELRANGE_4G                = (0b001 << 3),
  LSM9DS0_ACCELRANGE_6G                = (0b010 << 3),
  LSM9DS0_ACCELRANGE_8G                = (0b011 << 3),
  LSM9DS0_ACCELRANGE_16G               = (0b100 << 3)
} lsm9ds0AccelRange_t;

typedef enum
{
  LSM9DS0_ACCELDATARATE_POWERDOWN      = (0b0000 << 4),
  LSM9DS0_ACCELDATARATE_3_125HZ        = (0b0001 << 4),
  LSM9DS0_ACCELDATARATE_6_25HZ         = (0b0010 << 4),
  LSM9DS0_ACCELDATARATE_12_5HZ         = (0b0011 << 4),
  LSM9DS0_ACCELDATARATE_25HZ           = (0b0100 << 4),
  LSM9DS0_ACCELDATARATE_50HZ           = (0b0101 << 4),
  LSM9DS0_ACCELDATARATE_100HZ          = (0b0110 << 4),
  LSM9DS0_ACCELDATARATE_200HZ          = (0b0111 << 4),
  LSM9DS0_ACCELDATARATE_400HZ          = (0b1000 << 4),
  LSM9DS0_ACCELDATARATE_800HZ          = (0b1001 << 4),
  LSM9DS0_ACCELDATARATE_1600HZ         = (0b1010 << 4)
} lm9ds0AccelDataRate_t;

typedef enum
{
  LSM9DS0_MAGGAIN_2GAUSS               = (0b00 << 5),  // +/- 2 gauss
  LSM9DS0_MAGGAIN_4GAUSS               = (0b01 << 5),  // +/- 4 gauss
  LSM9DS0_MAGGAIN_8GAUSS               = (0b10 << 5),  // +/- 8 gauss
  LSM9DS0_MAGGAIN_12GAUSS              = (0b11 << 5)   // +/- 12 gauss
} lsm9ds0MagGain_t;

typedef enum
{
  LSM9DS0_MAGDATARATE_3_125HZ          = (0b000 << 2),
  LSM9DS0_MAGDATARATE_6_25HZ           = (0b001 << 2),
  LSM9DS0_MAGDATARATE_12_5HZ           = (0b010 << 2),
  LSM9DS0_MAGDATARATE_25HZ             = (0b011 << 2),
  LSM9DS0_MAGDATARATE_50HZ             = (0b100 << 2),
  LSM9DS0_MAGDATARATE_100HZ            = (0b101 << 2)
} lsm9ds0MagDataRate_t;

typedef enum
{
  LSM9DS0_GYROSCALE_245DPS             = (0b00 << 4),  // +/- 245 degrees per second rotation
  LSM9DS0_GYROSCALE_500DPS             = (0b01 << 4),  // +/- 500 degrees per second rotation
  LSM9DS0_GYROSCALE_2000DPS            = (0b10 << 4)   // +/- 2000 degrees per second rotation
} lsm9ds0GyroScale_t;


#ifdef __cplusplus
}
#endif

/* end of accel.h */
#endif

