//
// 12th Doctor's second Sonic Screwdriver
//
// - five RGB 'neopixel' LEDs with multiple effects
// - Sonic sounds by Direct Digital Synthesis and from stored 'wav' tables
// - synthesises the theme tune melody
// - Accelerometer/Gyro/Compass (on FLORA 9-DOF board) to handle gestures and effects
// - target hardware: Arduino Pro Mini, developed on Arduino Nano (the 'pro mini' with USB)
//   this board has an AtMega328 processor. Some stuff is specific for this processor!!!
//
// This software does not have a licence. Do not ask me for one. Just take it.
// It is free
// It is yours 
//
// TonyWilk
//   4th Dec 2017
//   mail@tonywilk.co.uk 
//   this may also be at http://tonywilk.no-ip.org/wardoctor when I get around to it.
//
#include <Arduino.h>
#include <avr/interrupt.h> 
#include <avr/pgmspace.h>
#include <accel.h>              // my simple accelerometer interface routines

#include "LED_neopixel.h"       // neopixel driving stuff is in here
#include "notes.h"              // definitions for musical notes
#include "themetune.h"          // the Dr. Who theme tune melody
#include "sounddata.h"          // and some converted .wav files...
#include "test.h"

#define LED_ON()  (PORTB |= 0x20)     // only used for timing debug
#define LED_OFF() (PORTB &= 0xdf)

void check_mem();                     // DEV: checks heap and stack
uint32_t mainLoopMicrosecs;
#define MAIN_LOOP_MICROSECS 10000     // pace main loop to 10mS

//-----------------------------
// Serial.print helper function
// - a real cut-down printf()
//
void Sprint( char *fmt, ... )
{
  char c;
  va_list args;
  va_start( args, fmt );
  while( (c=*fmt) != 0 ){
    switch( c )
    {
    case '%':
      c= *(++fmt);
      switch( c )
      {
      case 'd': Serial.print( va_arg(args,int) ); break;
      case 'f': Serial.print( va_arg(args,double) ); break;
      case 'h': Serial.print( va_arg(args,int), HEX ); break;
      case 'c': Serial.print( (char)va_arg(args,int) ); break;
      case 's': Serial.print( va_arg(args, char *) ); break;
      default:  break;
      }
      break;
    case '\\':
      c= *(++fmt);
      if( c == 'n' )
        Serial.println();
      else
        Serial.print( c );
      break; 
    default:
      Serial.print( c );
      break;
    }
    ++fmt;
  }
  va_end( args );
}


//----------------------------------------------------
// Sine wave lookup table...
//
byte wave[256]={
   128, 131, 134, 137, 140, 143, 146, 149, 152, 155, 158, 161, 164, 167, 170, 173, 
   176, 179, 182, 185, 187, 190, 193, 195, 198, 201, 203, 206, 208, 210, 213, 215, 
   217, 219, 222, 224, 226, 228, 230, 231, 233, 235, 236, 238, 240, 241, 242, 244, 
   245, 246, 247, 248, 249, 250, 251, 251, 252, 253, 253, 254, 254, 254, 254, 254, 
   255, 254, 254, 254, 254, 254, 253, 253, 252, 251, 251, 250, 249, 248, 247, 246, 
   245, 244, 242, 241, 240, 238, 236, 235, 233, 231, 230, 228, 226, 224, 222, 219, 
   217, 215, 213, 210, 208, 206, 203, 201, 198, 195, 193, 190, 187, 185, 182, 179, 
   176, 173, 170, 167, 164, 161, 158, 155, 152, 149, 146, 143, 140, 137, 134, 131, 
   128, 124, 121, 118, 115, 112, 109, 106, 103, 100, 97, 94, 91, 88, 85, 82, 
   79, 76, 73, 70, 68, 65, 62, 60, 57, 54, 52, 49, 47, 45, 42, 40, 
   38, 36, 33, 31, 29, 27, 25, 24, 22, 20, 19, 17, 15, 14, 13, 11, 
   10, 9, 8, 7, 6, 5, 4, 4, 3, 2, 2, 1, 1, 1, 1, 1, 
   1, 1, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 6, 7, 8, 9, 
   10, 11, 13, 14, 15, 17, 19, 20, 22, 24, 25, 27, 29, 31, 33, 36, 
   38, 40, 42, 45, 47, 49, 52, 54, 57, 60, 62, 65, 68, 70, 73, 76, 
   79, 82, 85, 88, 91, 94, 97, 100, 103, 106, 109, 112, 115, 118, 121, 124
};

//----------------------------------------------------
// 16-byte ring buffer for audio samples
//
uint8_t sampleBuffer[16];

uint8_t (*pGetSampleFunc)();    // pointer to current sample provider function
uint8_t bufOutIndex;            // index to next sample to output
uint8_t bufFillIndex;           // ring buffer refill index
uint8_t fastISR=0;              // Sound generation ISR speed switch

//----------------------------------------------------
// initSampleBuffer
// - prefills the 16-byte sample buffer on startup
//
void initSampleBuffer()
{
  // pre-fill buffer with samples
  for( int i=0; i<16; i++ ){
    sampleBuffer[i]= pGetSampleFunc();
  }
  bufOutIndex= 0;
  bufFillIndex= 0;  
}

//----------------------------------------------------
// sonic LEDs
// - leds[4] is the centre one
//
#define PIXELS 5
struct LED_PIXEL leds[ PIXELS ];        // colour and brightness of each LED
struct LED_RGB   pixelBuf[PIXELS];      // raw pixel buffer required by sendLeds()

int globalBrightness= 200;

//----------------------------------------------------
//
// We need to turn off the Arduino's Timer0 interrupt while
// the LED output is driven...
//
void timer0_off(){
  TIMSK0 &= ~_BV(TOIE0); // disable timer0 overflow interrupt
}
void timer0_on(){
  TIMSK0 |= _BV(TOIE0);  // enable timer0 overflow interrupt
}


//-------------------------------------------------------------------------------------
//                              Sonic sound generation
//-------------------------------------------------------------------------------------
// - uses simple Direct Digital Synthesis to generate the sonic waveform at c. 31.250KHz
// - the sound elements are:
//    a centre fequency of c. 3650Hz
//    FM modulated at c. 70Hz for a spread of 2500Hz to 4700Hz
//    then AM modulated at 130Hz - 170Hz  ?????????????
// - The control values for the elements are randomly adjusted in main-line code
//   to make the appropriate randomly wavering sound.
//
// - Notes on how DDS works:
//   There's a table of 256 sine values in wave[]
//   If we have an index into the wave table and, say, add 1 to it every time, we end up
//   extracting a sequence of values which is a sine wave.
//   Since we are doing this at a fixed frequency - the rate of the Timer2 interrupt
//   which is every 32uS - then by adding 1 to the index we would take 256 * 32uS
//   to generate one complete sine wave. This would be a total of 8,193uS which would be
//   an audio frequency of 122Hz.
//   To generate other frequencies, we make bigger steps in the wave table index.
//   For example, if we added 32 every time, we would step 0,32,64,96,128,160,192,224,0 (wraps at 256)
//   so it only takes 8 * 32uS = 256uS to output an approximation of a sine wave at 3,906Hz
//   This is a bit of a crap sinewave output with only 8 steps in it, but we are quite limited
//   in this processor only running at 16MHz. However, the result is not too bad when we filter
//   the output a bit.
//   Frequency resolution is helped by using a 16-bit phase value, the wave[] table is indexed
//   by the upper 8 bits. So to make 122Hz we would add 256 to phase for each sample.
//
// - Notes on the Sonic Sound generation: 
//   Examining example Sonic sounds with SoundForge or Audacity using FFT, you can find out that
//   the sound is mostle a 3650Hz tone which is 'spread' by frequency modulation to 2500Hz - 4700Hz
//   In the DDS code, we take the main tone index step value and add in a sine value for the
//   modulation - this makes the step value larger and smaller making a higher and lower frequencies
//   There is also an amplitude modulation which scales the output value according to the sine
//   value of the AM frequency.
//   With this implemention we do not have fine control over frequency generation, so we have to
//   fiddle about with some of the values to make a reasonable sonic warbling noise.
//   Some sets of values might suddlenly sound really awful if the numbers are multiples of each other
//   so the generation is synchronised instead of the main tone, FM and AM modulation all being out
//   of step with each other.
//   For this reason, odd numbers often work better than even ones... for example... if we wanted a
//   simple tone at around 1,000Hz we could increase the wave table index by 16 each time, this should
//   give us 16 steps to go thru the 256 sine entries which would take 16 * 64uS = 1,024uS to give
//   us audio at 976Hz. However, we would be evenly stepping through the sine table - we would 
//   always get indexes 0,16,32,48,64... and so on. A much better sound would be generated by stepping
//   by 15 (result would be 1,041Hz) and would use all the wavetable values: 0,15,30,45... then wrap
//   over to use 14,29,44,59... the next time thru and so on.
//
//   15350 = 3650Hz
//    4352 = 1000Hz
//     300 = 70Hz
//
// Using 16-bit phase rotor:
// at 31250Hz sampling, count of 65536 = 1 cycle so 1Hz must count to 65536 in 31250 samples
// i.e. it must step by 65536/31750 = 2.097152
//
#define Hz_TO_PHASE_STEP(hz) ((int)( (float)(hz) * 2.0097152f ))

// To make the sonic sound, we also want a slow beat between AM and FM modulations
// so we set AM = 185Hz, FM= 62Hz where 62*3 = 186Hz giving us ~ 1Hz beat
//
#define FREQ_MID    Hz_TO_PHASE_STEP( 3650 )    // main centre frequency 3650Hz
#define FREQ_FM     Hz_TO_PHASE_STEP(  128 )    // FM modulation frequency  (c. 60Hz)
#define FREQ_AM     Hz_TO_PHASE_STEP(  137 )    // AM modulation frequency  (185Hz)
#define FM_SCALE    12       // scaling factor of FM modulation depth

//
// - Later addition of an accelerometer will adjust these values
//   based on movement of the sonic.
//
uint16_t myint;

uint16_t mainPhase= 0;        // phase rotor for main frequency
uint16_t mainMid= FREQ_MID;   // centre freq of 3650Hz by adding this value to phase
uint16_t mainAdd; 

uint16_t fmPhase= 0;          // fm modulation phase rotor (c. 70Hz)
uint16_t fmAdd= FREQ_FM;
uint16_t fmSine= 0;  
uint16_t _fmScale;            // scaling factor of FM modulation depth

// Set fm scaling factor to a value 1..15  ???
//
void set_fmScale( int val )
{
//  _fmScale= 256 / (val*2);     // this factor is pre-divided to save time in the interrupt
    _fmScale= val;
}

uint16_t amPhase=0;           // phase rotor for Amplitude Modulation
uint16_t amAdd= FREQ_AM;
uint8_t  amSine= 0;
uint8_t  val;
uint8_t  volume= 255;         // overall sound volume

//---------------------------------------
// preprocessTune
// - convert tune table from
//     frequency, duration_in_millisecs
// to
//     phaseIncrement, duration_in_samples
//
// We keep the melody[] table as simple Note Frequency, Node duration in mS
// - this is handy for a simple player using the Arduino tone() function
// - to use the table with our DDS synth, it's simpler to pre-process the table
//   into friendlier values.
//
// TEMPO is just a number chosen so the duration comes out as no. of samples for the note to play
//
void preprocessTune( int *tune ){
  uint16_t i= 0;
  uint16_t n,d,decayBits;
  while( tune[i] != NOTE_END ){
    n= tune[i];
    decayBits= n & NOTE_DECAY_MASK;
    n= n & ~NOTE_DECAY_MASK;
    d= tune[i+1];
    if( n > NOTE_CMD_END ){
      n= Hz_TO_PHASE_STEP( n );       // it's a real note, convert frequency to phase increment
    }
    d= d*32*TEMPO;                    // time in samples for 32KHz mode
    tune[i++]= n | decayBits;
    tune[i++]= d;
  }
}
// tune playing variables
uint16_t tuneIndex=0;
uint16_t noteDuration=0;
uint16_t nextPhaseIncr=0;
uint16_t notePhaseIncr=0;
uint16_t notePhase, notePhase2;
uint16_t wavPtr= 0;

uint16_t auxFadeCount,auxFadeDivider;     // levels used for LEDs while playing melody
uint8_t  auxLevel;

// General notes of generating sounds...
// - Timer 2 ISR is called 31,250 times per second, it writes an 8-bit sample to the PWM output
// - samples are generated by the currently-selected sample function via the function pointer: pSampleFunc
// - there are a number of different sample generators:
//    getSonicSample()    generates sonic sound effect by Direct Digital Synthesis
//    getTuneSample()     plays the Dr Who theme melody by synthesising tones from a table
//    getWav32Sample()    plays a "wav file" stored in program memory ( at 32KHz sampling)
//    getWav8Sample()     plays a "wav file" stored in program memory ( at 8KHz sampling)
//    getWav4Sample()     plays a "wav file" stored in program memory ( at 4KHz sampling)
//
// - setSoundMode() sets the sample generator function and does initialisation stuff.

extern "C" {

  //
  // These are "C" routines 'cos I was trying to write the Timer2 ISR in assembler only
  // where a straight 'call' instruction wants a C function.
  //
  void    setSoundMode( int mode );
  uint8_t getSonicSample() __attribute__((used));      // get sample from sonic sound synth
  uint8_t getTestSample()   __attribute__((used));
  uint8_t getWavSample()   __attribute__((used));
  uint8_t getTuneSample()  __attribute__((used)) ;
  uint8_t getSilentSample() __attribute__((used)) ;
  uint8_t getSynth2Sample() __attribute__((used)); 
  
  void    initSonicSample();
  void    initTestSample();
  void    initWavSample();
  void    initTuneSample();
  void    initSynth2Sample();

  void setSoundMode( int mode ){
    cli();    // interrupts off while we change pointers
    switch( mode ){
    case 1:   pGetSampleFunc= getSonicSample;   initSonicSample();  break;
    case 2:   pGetSampleFunc= getTuneSample;    initTuneSample();   break;
    case 3:   pGetSampleFunc= getTestSample;    initTestSample();   break;
    case 4:   pGetSampleFunc= getWavSample;     initWavSample();    break;
    case 5:   pGetSampleFunc= getSynth2Sample;  initSynth2Sample(); break;
    default:  pGetSampleFunc= getSilentSample;  break;
    }
    sei();    // Enable interrupts to generate waveform!
  }

  //------------------------------------------
  // sample function to just return silence...
  // - sets PWM to 128 to maintain DC level
  //
  uint8_t getSilentSample(){
    return 128;
  }
  
  //------------------------------------------
  // Test to get a TEST WAV sample... (data is at 4Khz)
  //
  void initTestSample()
  {
    wavPtr= 0;
  }
  uint8_t getTestSample()
  {
    static uint8_t div=8;
    if( --div == 0 ){
      div= 8;
      if( ++wavPtr >= TEST_LENGTH )
        wavPtr= 0;
    }
    return pgm_read_byte( testdata + wavPtr );
  }

  //------------------------------------------
  // Test to get a WAV sample... (data is at 8KHz)
  //
  void initWavSample()
  {
    wavPtr= 0;
  }
  uint8_t getWavSample()
  {
    static uint8_t div=4;
    if( --div == 0 ){
      div= 4;
      if( ++wavPtr >= SAMPLE_LENGTH )
        wavPtr= 0;
    }
    return pgm_read_byte( sounddata + wavPtr );
  }

  //------------------------------------------
  // Get a sample for playing the theme tune melody
  // - melody[] contains int pairs of: note_phase_increment, duration_in_samples
  // 
  uint16_t noteFadeDivider= 0;
  uint16_t noteFadeCount= 0;
  uint8_t  noteVolume;
  uint16_t decayType, decayShift;

  void initTuneSample()
  {
    tuneIndex= 0;
    notePhaseIncr= NOTE_REST;
    noteDuration= 0;
  }
  uint8_t getTuneSample()
  {
    // melody
    if( noteDuration == 0 ){
      if( melody[tuneIndex] == NOTE_END ){
        tuneIndex=0;
      }
      // get next note;
      nextPhaseIncr= melody[tuneIndex++];
      noteDuration= melody[tuneIndex++];

      // note volume decay
      decayType= nextPhaseIncr & NOTE_DECAY_MASK;   // get upper 2 decay bits
      nextPhaseIncr &= ~NOTE_DECAY_MASK;            // keep lower bits as actual phase increment
      decayShift= (decayType >> 14) + 5;            // 1,2,3 -> 6,7,8
      if( decayType == 0  ){
        noteFadeDivider= 0;                         // no note decay
        noteVolume= volume;
        auxLevel= 255;
        auxFadeDivider= (noteDuration >> 7);        // always decay aux
      }else
      {
        noteFadeDivider= (noteDuration >> decayShift);  // calc divider for decrementing volume
        noteFadeCount= noteFadeDivider;             // set initial division count
        auxFadeDivider= noteFadeDivider;            // decay aux same as note
      }
      auxFadeCount= auxFadeDivider;

      if( notePhaseIncr == NOTE_REST )    // after rest, start note immediate
        notePhaseIncr= nextPhaseIncr;   
      else
      if( nextPhaseIncr == NOTE_HOLD )    // hold existing note
        nextPhaseIncr= notePhaseIncr;
      else
      if( nextPhaseIncr == NOTE_REST ){    // set rest immediate
        notePhaseIncr= NOTE_REST;
      }
                                          // else we slide to the new note
    }
    if( notePhaseIncr != nextPhaseIncr )
    { 
      // ramp notePhaseIncr up/down to reach nextPhaseIncr
      if( notePhaseIncr < nextPhaseIncr )
      {
        notePhaseIncr+= 3; 
        if( notePhaseIncr > nextPhaseIncr )
          notePhaseIncr= nextPhaseIncr;
      }else{
        notePhaseIncr-= 3; 
        if( notePhaseIncr < nextPhaseIncr )
          notePhaseIncr= nextPhaseIncr;
      }
    }      
    noteDuration--;

    if( notePhaseIncr != NOTE_REST ){
      
      notePhase += notePhaseIncr;             // main note frequency
      notePhase2 += (notePhaseIncr<<1) - 1;   // overtone and a tiny bit of vibrato
  
      val= wave[notePhase>>8]>>1;         
      val+= wave[notePhase2>>8]>>2;           // adds "harmonics" and vib

      if( noteFadeDivider != 0 )
      {
        if( --noteFadeCount == 0 )            // decrement volume on each division countdown
        {
          noteFadeCount= noteFadeDivider;
          if( noteVolume > 0 )
            --noteVolume;
        }
      }
      // overall volume
      val = ((((int)val - 128) * (int)noteVolume) - 32768) >> 8;

      // auxLevel fades for each note... we use this for driving the LEDs
      if( --auxFadeCount == 0 ){
        auxFadeCount= auxFadeDivider;
        if( auxLevel > 0 )
          --auxLevel;
      }
      
    }else{
      // leave PWM at current level during REST
      auxLevel= 0;
    }

    return val;
  }

  //------------------------------------------
  // Synth 2 - testing....
  //
  int wav1, wav2;
  uint16_t tmp, rmPhase, rmAdd, rmSine;
  void initSynth2Sample()
  {
    fmPhase= amPhase= 0;
  }
  inline uint8_t getSynth2Sample()
  {    
      fmPhase+= fmAdd;
      fmSine= wave[fmPhase>>8];             // get FM modulation sine value 0..255
      
      amPhase+= amAdd;
      amSine= wave[amPhase>>8];

      rmPhase+= rmAdd;
      rmSine= wave[rmPhase>>8];
/*      
      if( fmSine > 127 ){
        // add fm offset
        fmSine= (fmSine - 128) * _fmScale;  // note: result is 16-bit
        mainAdd= mainMid + fmSine;
      }else{
        // subtract fm offset
        fmSine= (127 - fmSine) * _fmScale;
        mainAdd= mainMid - fmSine;
      }
*/
      mainAdd= mainMid;  //DEBUG: uncomment this to turn off FM

//amPhase= fmPhase << 1;    
//      tmp= amPhase + (fmPhase << 1);   
//     amSine= wave[tmp>>8];
//amSine= (amPhase & 0x8000) ? 250 : 0;

      // scaling of AM depth is not worth it
      // - only real difference is AM on or off
      //amSine= (((int)amSine * (int)amScale) >> 8) + (256 - amScale);

      mainPhase += mainAdd;

      val = wave[mainPhase>>8]; // get main output value

      wav1= (int)val - 127;
      wav2= (int)(rmSine) - 127;  // ring modulator 
      // ring modulate + add in fundamental again...
      val= (uint16_t)((wav1 * wav2) + (wav1 * 127) + 32760) >> 8;

      wav1= (int)val - 127;
      wav2= (int)(fmSine) - 127;  // ring modulate again with fmPhase 
      val= (uint16_t)((wav1 * wav2) + (wav1 * 127) + 32760) >> 8;


      // AM modulate
      // (DEBUG: comment this out to disable AM)
//      val = ((((int)val - 128) * (int)amSine) - 32768) >> 8;

      // overall volume
      val = (val * volume ) >> 8;
  
      return val;
  }    

  //------------------------------------------------
  // Sonic sound synth
  //
  void initSonicSample()
  {
    mainMid= FREQ_MID;
    fmAdd= FREQ_FM;
    amAdd= FREQ_AM;
    set_fmScale( FM_SCALE );
    fmPhase= amPhase= 0;
  }
  inline uint8_t getSonicSample()
  {    
      fmPhase+= fmAdd;
      fmSine= wave[fmPhase>>8];             // get FM modulation sine value 0..255
      
      if( fmSine > 127 ){
        // add fm offset
        fmSine= (fmSine - 128) * _fmScale;  // note: result is 16-bit
        mainAdd= mainMid + fmSine;
      }else{
        // subtract fm offset
        fmSine= (127 - fmSine) * _fmScale;
        mainAdd= mainMid - fmSine;
      }

      //mainAdd= mainMid;  //DEBUG: uncomment this to turn off FM

      amPhase+= amAdd;
      amSine= wave[amPhase>>8];
//amPhase= fmPhase << 1;    
//      tmp= amPhase + (fmPhase << 1);   
//     amSine= wave[tmp>>8];
//amSine= (amPhase & 0x8000) ? 250 : 0;

      // scaling of AM depth is not worth it
      // - only real difference is AM on or off
      //amSine= (((int)amSine * (int)amScale) >> 8) + (256 - amScale);

      mainPhase += mainAdd;
      val = wave[mainPhase>>8]; // get main output value

      //wav1= (int)val - 127;
      //wav2= (int)amSine - 127;  // ring modulator 
      //val = (uint16_t)((wav1 * wav2)+16256) >> 7;     // ring modulate ?

      
      // AM modulate
      // (DEBUG: comment this out to disable AM)
      val = ((((int)val - 128) * (int)amSine) - 32768) >> 8;

      // overall volume
      val = (val * volume ) >> 8;
  
      return val;
  }    

  //------------------------------------------
  // refill the sample buffer
  // - adds one sample
  // - then adds another if the buffer is not full
  // - called from the Timer2 ISR
  //
  void refillBuffer() __attribute__((used));
  void refillBuffer()
  {
    // refill sample buffer
    sampleBuffer[bufFillIndex]= pGetSampleFunc();
    bufFillIndex= (bufFillIndex + 1 ) & 0x0f;
    
    // catch-up fill sample buffer if not full
    if( bufFillIndex != bufOutIndex){
      sampleBuffer[bufFillIndex]= pGetSampleFunc();
      bufFillIndex= (bufFillIndex + 1 ) & 0x0f;
    }  

  }
}// extern "C"



//--------------------
// Timer 2 ISR
// - Called every time TCNT2 = OCR2A
// - this is set at c. 31.250KHz to write samples to the fast PWM for audio output 
//
// BIG complication...
// - the RGB LED string is written in main-line code and takes approx. 300uS
//   if interrupts are disabled, we get objectionable clicks in the audio.
//   interrupts are disabled only for the time-critical high periods of the LED drive waveform,
//   however, the low periods will be interrupted by this ISR approx. 5 times
//   If this ISR takes more than c. 6uS, then the LED drive is corrupted (LED's see a reset)
//   So... we maintain a buffer of samples which can be used in 'fastISR' mode
//
// - in normal mode, the ISR takes one sample from the buffer and write it to the PWM
//   then calculates a sample to refill the buffer. If the buffer is not full, a second
//   sample is calculated and buffered. (we only calc two to limit time in the ISR)
// - in fast mode, the ISR just takes a sample from the buffer.
//
//
uint8_t xxx;
ISR(TIMER2_COMPA_vect) { // Called when TCNT2 == OCR2A
  
  //LED_ON();    // DEBUG - led on

  OCR1AL= sampleBuffer[bufOutIndex];    // Update the PWM output
  bufOutIndex= (bufOutIndex + 1) & 0x0f;  

  if( fastISR == 0 )          // if not in 'fastISR' mode
    refillBuffer();           // - we refill the buffer

  //LED_OFF();    // DEBUG - led off
}



//-------------------------------------------------------------------------------
//                                    setup
//-------------------------------------------------------------------------------
uint8_t soundMode= 1;     // default startup sound mode
    
void setup() {
    uint32_t color;

  // initialize digital pin LED_BUILTIN as an output.
  pinMode(LED_BUILTIN, OUTPUT);
  LED_ON();
  LED_OFF();
  LED_ON();

  // *DEBUG* initialize serial communication at 9600 bits per second:
  Serial.begin(19200);

  Sprint("This is a test\n");
  Sprint("int:%d, float:%f, hex:%h, char %c, end\n", 42, 123.45, 0xFACE, 'x' );

  //check_mem();

  // initialise accelerometer
  for( int i=0; i<4; i++ ){
    delayMicroseconds( 500 );
    if( accel_init() == I2C_OK ){
      Serial.println(F("Accel Ok "));
      break;
    }else{
      Serial.println(F("Accel FAIL"));
    }
  }

  cli();    // interrupts off while we set stuff up

  setSoundMode( soundMode );  // set default sound mode (MUST do this first!)
  initSampleBuffer();         // initialise waveform output buffer
  set_fmScale( FM_SCALE );    // set default fm scaling (modulation depth)
  preprocessTune( melody );   // preprocess melody ready for sample generation


  bitSet( DDRD, 7 );   // D7 output
  PORTD |= 0x80;

  ledSetup();

  // Set timer1 for 8-bit fast PWM output 
  pinMode(9, OUTPUT); // Make timer’s PWM pin an output

  TCCR1B = (1 << CS10);     // Set prescaler to full 16MHz
  TCCR1A |= (1 << COM1A1);  // Pin low when TCNT1=OCR1A
  TCCR1A |= (1 << WGM10);   // Use 8-bit fast PWM mode
  TCCR1B |= (1 << WGM12);

  // Set up timer2 to call ISR 
  TCCR2A = 0x02;                        // WGM21=1, CTC mode 
  TCCR2B = (1 << FOC2A) | (1 << CS21);  // Force o/p compare A, Set prescaler to divide by 8
  TIMSK2 = (1 << OCIE2A);               // Call ISR when TCNT2 = OCRA2
  OCR2A = 63;                           // Set frequency of generated wave:
                                        //  63= 32uS = 31.250KHz sample rate
  LED_OFF();
  
  sei();      // Enable interrupts to generate waveform!

  mainLoopMicrosecs= micros();

  XYZ magRaw;
  setMagCalMode( MAG_CAL_INIT );

/*
  Serial.println("Calibrate Magnetometer......");
  while( (micros() - mainLoopMicrosecs) < 15000000 )
  {
    readMag( &magRaw );
    processMagCal( &magRaw );
  }
  setMagCalMode( MAG_CAL_STOP );
  getMagCal( &magRaw );
  Sprint( "calX: %d, calY: %d, calZ: %d\n", 
    (int)magRaw.X.val, (int)magRaw.Y.val, (int)magRaw.Z.val ); 
*/

  mainLoopMicrosecs= micros();
}

//-------------------------------------------------------------------------------

uint8_t phaseAdd= 2;      // adder for foreground phase angle adjusts
uint8_t phase=0;          // foreground phase angle
uint8_t ledMode= 0;       // current LED effect mode
uint8_t modeTimer= 0;     // timer for each display mode
int soundRandTime= 1;

void devAdjustParams();

VECXYZ axisRaw, axisAccel, axisDegrees;


//-------------------------------------------------------------------------------
//                                    loop
//-------------------------------------------------------------------------------
volatile uint8_t abyte;

void loop() { 
  int j, k;
  static int i=100;
  static uint8_t secsTimer= 0;     // *debug*
  static int deadtime=50;
  uint8_t sinval;
  int heading;
  XYZ magRaw;

  //check_mem();      // print heap and stack pointers

  processAccel( &axisRaw, &axisAccel, &axisDegrees );     // get and process accelerations

/*
  readMag( &magRaw );
  heading= calcHeading( &magRaw );
  Sprint("X: %d, Y: %d, Z: %d, heading: %d",
    (int)(magRaw.X.val),(int)(magRaw.Y.val),(int)(magRaw.Z.val), heading );
  delay(100);
*/
  // Sonic sound parameter adjustments...
  //
  if( soundMode == 1 )    // adjust synth Sonic by accel angles...
  {
    cli();    // interrupts off
    mainMid= FREQ_MID + (axisDegrees.x << 4);
    fmAdd= FREQ_FM + (axisDegrees.z >> 2);
    sei();    // Enable interrupts
  }
  if( ++secsTimer > 20 )    // 20*10mS= 200mS
  {
    secsTimer= 0;
    Sprint( "X= %d, Y= %d, Z=%d\n", axisDegrees.x, axisDegrees.y, axisDegrees.z);
  }
  
/*  
  phase += phaseAdd;
  sinval= wave[phase];
  j= (int)sinval << 3;
  //Serial.println(j);
  mainMid= FREQ_MID + j;      // slow frequency waver

  mainMid= FREQ_MID - axisDegrees.x << 4;
//  phaseAdd= axisDegrees.x >> 3;

  fmAdd = FREQ_FM - (axisDegrees.x >> 2);
*/  
  //devAdjustParams();    // *DEV* adjust sound parameters
/*
  if( --soundRandTime <= 0 )
  {
    soundRandTime= random( 40, 100 );    // 5000 .. 1200mS

    //fmScale= FM_SCALE + random(-1,2);
    amAdd= FREQ_AM + random(-3,3);
    fmAdd= FREQ_FM + random (-2,2);
    phaseAdd= random( 1,3 );
  }
*/
/*
  if( ++secsTimer > 130 ){
    secsTimer= 0;
    if( ++modeTimer > 5 ){
      modeTimer= 0; 
      ledMode= (ledMode + 1) & 0x03;
    }
  }
*/  

  // Detect gestures and switch modes
  //
  if( deadtime > 0 ){
    --deadtime;
  }else{ 
    deadtime= 30;
    if( (axisAccel.z < -500) && (axisDegrees.z > 30) ){
      Sprint( "flick right\n" );
      ledMode= (ledMode + 1) & 0x03;
    }else  if( (axisAccel.z > 500) && (axisDegrees.z < -30) ){
      Sprint( "flick left\n" );
      soundMode++;
      if( soundMode > 5 )
        soundMode= 0;
      setSoundMode( soundMode );
      
    }else  if( axisAccel.x > 1500 ){  // reset on thrust forward
      Sprint( "thrust forward\n");
      soundMode= 1;
      setSoundMode( soundMode );
      ledMode= 0;
    }else  if( axisAccel.y > 2000 ){
      Sprint( "drop down\n" );
    }else{
      deadtime=0;
    }
  }

  doLedEffect( ledMode );     // do a step of current LED effect

  //LED_ON();
  
  // pace main loop 
  while( (micros() - mainLoopMicrosecs) < MAIN_LOOP_MICROSECS )
    ;
  mainLoopMicrosecs= micros();
  
  //LED_OFF();
}

//--------------------------------------------------
// doLedEffect
// - processes a step of the current LED effect
// - outputs data to LEDs 
// - called from the main loop() approx every 10mS
// - some effect steps are timed by counting (increments of ~ 10mS)
// - some effects step thru the sine wave[] table to vary LED brightness
//   where the speed is determined by the phase step
//
static uint8_t lastMode=255;
static uint8_t ledDelay;
static uint8_t ledPulse;      // LED effect parameters
static uint8_t ledState;      // current LED sequence state
uint8_t initMode;


void doLedEffect( uint8_t ledMode )
{
  uint16_t j, k;
  uint8_t c;

  initMode= 0;
  if( ledMode != lastMode ){
    lastMode= ledMode;
    initMode= 1;                // initialise on mode change
  }
  
  switch( ledMode )
  {
  case 0:
    // Rotating blue
    if( initMode ){
      ledDelay= 255;
      ledState= 0;
      ledPulse= 0;
    }
    if( ++ledDelay >= 10 ){
      ledDelay= 0;
      ledState= (ledState+1) & 0x03;    // 0,1,2,3
      for( j=0; j<4; j++ ){
        leds[j].colour= (j==ledState) ? BLACK : BLUE;
        leds[j].bright= 128;
      }  
      leds[4].colour= BLUE;
      leds[4].bright= 200;
    }
    ledPulse+= 10;
    leds[4].colour= BLUE;
    leds[4].bright= wave[ledPulse];
    break;
    
  case 1:
    // pulsing green
    if( initMode ){
      ledPulse= 0;
    }
    ledPulse += 6;
    for( j=0; j<5; j++ ){
      leds[j].colour= GREEN;
      leds[j].bright= (wave[ledPulse]>>1) + 100;
    }
    leds[4].bright= 255 - leds[4].bright;
    break;
    
  case 2:
    // alternating RED
    if( initMode ){
      ledPulse= 0;
    }
    ledPulse += 6;
    ledState= wave[ ledPulse ];
    for( j=0; j<5; j++ ){
      leds[j].colour= RED;
      leds[j].bright= ledState;
      ledState= 255 - ledState;
    }
    leds[4].bright= (ledState > 128) ? ledState : 255-ledState;   // pulse middle led
    break;
    
  default:
    // rainbow + pulsing middle led
    if( initMode ){
      leds[0].colour= 0x00;
      leds[1].colour= 0x10;
      leds[2].colour= 0x20;
      leds[3].colour= 0x30;
      ledPulse= 0;
    }
    for( j=0; j<PIXELS; j++ ){
      leds[j].bright= 128;
      leds[j].colour += 2;
      if( leds[j].colour >= 240 )
        leds[j].colour -= 240;
    }
    leds[4].colour= PURPLE;
    ledPulse += 4;
    leds[4].bright= wave[ ledPulse ];

    // modulate middle LED with melody...
    //
    j= notePhaseIncr >> 5;              // highest note: E7  2637Hz = 5442 phase step...
    if( j == 0 ){
      leds[4].colour= 0;
      leds[4].bright= 32;
    }else{
      // fiddle about to get a number we can use   was: <60
      // to index the colour table so it looks nice 
      j= (j < 30) ? 0 : j-30;
      j= j + (j >> 1);
      if( j > 223 )           // limit to rainbow part of palette
        j=223;
      leds[4].colour= j; 
      leds[4].bright= auxLevel;
    }
    
    
    break;
  }

  // LED output
  //
  sendLeds( leds, pixelBuf, globalBrightness, PIXELS );
}

//--------------------------------------------------
// *DEV*
// - wired a 10K potentiometer into analog channel 0
// - used to investigate varying parameters for sound effect

//#define FREQ_MID    Hz_TO_PHASE_STEP( 3650 )    // main centre frequency 3650Hz
//#define FREQ_FM     Hz_TO_PHASE_STEP(  128 )    // FM modulation frequency  (c. 60Hz)
//#define FREQ_AM     Hz_TO_PHASE_STEP(  137 )    // AM modulation frequency  (185Hz)
//#define FM_SCALE    18       // scaling factor of FM modulation depth

#define potPin 0

float mainFreq= 3650.0f;
float beatFreq= 3.5f;
float fmFreq= 118.0f;
float amFreq= 122.0f;

// Set FM frequency and adjust AM to maintain same beat frequency
//
void setFM_adjAM( float Hz )
{
  fmFreq= Hz;
  amFreq= fmFreq + beatFreq;
  fmAdd= Hz_TO_PHASE_STEP( fmFreq );       // frequency modulation rate
  amAdd= Hz_TO_PHASE_STEP( amFreq );       // amplitude modulation rate
}

// Set AM frequency and adjust FM to maintain same beat frequency
//
void setAM_adjFM( float Hz )
{
  amFreq= Hz;
  fmFreq= amFreq - beatFreq;
  fmAdd= Hz_TO_PHASE_STEP( fmFreq );       // frequency modulation rate
  amAdd= Hz_TO_PHASE_STEP( amFreq );       // amplitude modulation rate
}

// set beat frequency
// - this is the difference between AM and FM modulation to get a beat
// - resets FM and adjusts AM for this beat frequency
//
void setBeatFreq( float Hz )
{
  beatFreq= Hz;  
  setFM_adjAM( fmFreq );
}

// a range from 0..255 to produce a sonic-like range of sounds
// - 0= slow, low freq, 255=warbly high pitch
//
void setSonicA( uint8_t val )
{
   setBeatFreq( (float)val /  16.0f + 1.0f );    // set beat 1..16Hz
   setFM_adjAM( (float)val + 50.0f );           // FM 50..300Hz
}

void devAdjustParams()
{
  int i, analog, main, fm, am;
  static uint16_t ph;

  analog= analogRead( potPin );

  //set_fmScale( (analog >> 2) +1 );    // 8=thin whistly, 12=sonicy, 20=raspy, 36=deep raspy, 40=zizzy, 100=wierd
  set_fmScale( 12 );

//  mainAdd= Hz_TO_PHASE_STEP( mainFreq );    // main middle frequency
//  fmAdd= Hz_TO_PHASE_STEP( fmFreq );       // frequency modulation rate
//  amAdd= Hz_TO_PHASE_STEP( amFreq );       // ring modulator input

  //setAM_adjFM( (float)analog );
  //setBeatFreq( (float)analog /10.0f );

//  setSonicA( analog >> 2 );           // <--- this works Ok
  i= analog << 2;
  amAdd= Hz_TO_PHASE_STEP( i );
  fmAdd= Hz_TO_PHASE_STEP( 1730 );
  rmAdd= ( i );
  mainMid= Hz_TO_PHASE_STEP( 750 );

  ph= ph - i;
  //volume= wave[ph & 0x00ff];

  //Sprint( "%d", i );
  
/* 
  Sprint( "%f, main= %dHz, beat= %dHz, fm=%d, am=%d, fmScale=%d\n",
    (float)analog/10.0f,mainFreq,beatFreq,fmFreq,amFreq,_fmScale); 
*/
//  mainMid= (analog <<6);      // 8,000 .. 25,000 are useable
//  Sprint("mainMid= %d\n", mainMid);

//  amAdd= (analog >>1) + 250;
//  Sprint("amAdd= %d\n", amAdd );
  
  //analog= analog - 32;    // -32 .. 32
  //mainMid= 60+analog;     //Good - main frequency adjust 60 +-32
  //fmAdd= analog;          // 0..4 whistly, 5..10 warbly, 16..18 sonic sound, 32=ok, >32 not so good
  //amAdd= analog;          // not so useful (tried 0..64)
  //set_fmScale( analog );  // Useful effects 0..15
  //volume= analog;         // values 0..255 work fine
  //phaseAdd= analog;       // useful values 0..31    (2=nice sonic sound)
}


//---------------------------------------------------------------------------------
//                        Routines used in development 
//---------------------------------------------------------------------------------
// toHex
// - convert value to two hex characters
// - the usual Arduino functions output e.g. "0" instead of "00"
//
char *toHex( uint8_t c )
{
  const static char *hex="0123456789ABCDEF";
  static char toHexBuf[3];
  toHexBuf[0]= hex[c>>4];
  toHexBuf[1]= hex[c & 0x0f];
  toHexBuf[2]=0;
  return toHexBuf;
}

void check_mem() {
  uint8_t * heapptr, * stackptr; 
  stackptr = (uint8_t *)malloc(4);  // use stackptr temporarily
  heapptr = stackptr;               // save value of heap pointer
  free(stackptr);                   // free up the memory again (sets stackptr to 0)
  stackptr =  (uint8_t *)(SP);      // save value of stack pointer

  Sprint("Heap: %h,  Stack: %h\n",(uint16_t)heapptr, (uint16_t)stackptr >> 8 );
}

/********** not needed any more...

//----------------------------------------------------------
// SerialPrintColours
// - prints out the 8-bit pallette colours as an HTML table to the serial monitor
//   so we can view it in a browser for reference
//
void SerialPrintColours()
{
  int row, col, index;

  Serial.println("<html><head></head><body><h2>Colour Pallette</h2><table>");
  for( row=0; row<16; row++ ){
    Serial.println("<tr>");
    for( col=0; col<16; col++ ){
      index= col+row*16;
      Serial.print("<td bgcolor='#"); 
      Serial.print( toHex( pgm_read_byte_near(pal_red + index)) );  
      Serial.print( toHex( pgm_read_byte_near(pal_green + index)) );
      Serial.print( toHex( pgm_read_byte_near(pal_blue + index)) );  
      Serial.print("'>");
      Serial.print(index,DEC);
        Serial.print("<br>#");
        Serial.print( toHex( pgm_read_byte_near(pal_red + index)) );  
        Serial.print( toHex( pgm_read_byte_near(pal_green + index)) );
        Serial.print( toHex( pgm_read_byte_near(pal_blue + index)) );  
      Serial.print("</td>");
    }
    Serial.println("</tr>");
  }
  Serial.println("</table></body></html>");
}

//-----------------------------------------------------------------
// 
//  The following routines were used to generate the palette tables
//  fillPallette() filled the arrays palR,palG,palB
//  then SerialPrintPal() printed the tables to the serial monitor.
// 
//-----------------------------------------------------------------
// HSL_to_RGB() based on:
//
// A Fast HSL-to-RGB Transform
// by Ken Fishkin
// from "Graphics Gems", Academic Press, 1990
//
void HSL_to_RGB( double h, double sl, double l, int *ir, int *ig,int *ib)
{
  double r,g,b;
    double v;

  if( h >= 1.0 )
    h= 0.9999;

    v = (l <= 0.5) ? (l * (1.0 + sl)) : (l + sl - l * sl);
    if (v <= 0) {
    r = g = b = 0.0;
    } else {
    double m;
    double sv;
    int sextant;
    double fract, vsf, mid1, mid2;

    m = l + l - v;
    sv = (v - m ) / v;
    h *= 6.0;
    sextant = (int)h; 
    fract = h - sextant;
    vsf = v * sv * fract;
    mid1 = m + vsf;
    mid2 = v - vsf;
    switch (sextant) {
      case 0: r = v; g = mid1; b = m; break;
      case 1: r = mid2; g = v; b = m; break;
      case 2: r = m; g = v; b = mid1; break;
      case 3: r = m; g = mid2; b = v; break;
      case 4: r = mid1; g = m; b = v; break;
      case 5: r = v; g = m; b = mid2; break;
    }
    }
  if( r > 1.0 ) r= 1.0;
  if( g > 1.0 ) g= 1.0;
  if( b > 1.0 ) b= 1.0;

  *ir= (int)(r*255.0);
  *ig= (int)(g*255.0);
  *ib= (int)(b*255.0);
}
uint8_t palR[256];
uint8_t palG[256];
uint8_t palB[256];

void fillPallette()
{
  double hue, sat, lite;
  int r, g, b;
  sat= 1.0;
  lite= 0.5;
  for( int i=0; i<240; i++ )
  {
    hue= (double)i/240.0;
    HSL_to_RGB( hue, sat, lite, &r, &g, &b );
    palR[i]= r;
    palG[i]= g;
    palB[i]= b;
  }
  
  for( int i=0; i<14; i++ )
  {
    hue= (double)i/14.0;
    lite= 0.8;
    HSL_to_RGB( hue, sat, lite, &r, &g, &b );
    palR[240+i]= r;
    palG[240+i]= g;
    palB[240+i]= b;
  }
   
}

void SerialPrintPal()
{
  int row, col, index;

  Serial.println("// Red");
  for( row=0; row<16; row++ ){
    for( col=0; col<16; col++ ){
      index= col+row*16;
      Serial.print("0x" );
      Serial.print( toHex( palR[index] ));
      Serial.print( ", ");
    }
    Serial.println("");
  }
  Serial.println("// Green");
  for( row=0; row<16; row++ ){
    for( col=0; col<16; col++ ){
      index= col+row*16;
      Serial.print("0x" );
      Serial.print( toHex( palG[index] ));
      Serial.print( ", ");
    }
    Serial.println("");
  }
  Serial.println("// Blue");
  for( row=0; row<16; row++ ){
    for( col=0; col<16; col++ ){
      index= col+row*16;
      Serial.print("0x" );
      Serial.print( toHex( palB[index] ));
      Serial.print( ", ");
    }
    Serial.println("");
  }
}

//-----------------------------------------------------------------
// Sine wave parameters 
#define PI2 6.283185 // 2*PI saves calculation later
#define AMP 127 // Scaling factor for sine wave
#define OFFSET 128 // Offset shifts wave to all >0 values
#define LENGTH 256 // Length of the wave lookup table
byte wave[LENGTH]; // Storage for waveform

void makeSinewaveTable(){
  // Populate the waveform table with a sine wave
  for (int i=0; i<LENGTH; i++) { // Step across wave table
    float v = (AMP*sin((PI2/LENGTH)*i)); // Compute value
    wave[i] = int(v+OFFSET); // Store value as integer
  }

  // Print waveform table to serial
  //
  Serial.println("byte wave[256]={");
  for( int row=0; row<16; row++ ){
    Serial.print("   ");
    for( int col=0; col<16; col++ ){
      Serial.print( wave[row*16+col] );
      Serial.print(", ");
    }
    Serial.println("");
  }
  Serial.println("};");
}
**********/

